#include "serialComm.h"
#include "serialComm_LL.h"
#include "stm32f4xx_nucleo.h"

#include "FreeRTOS.h"
#include "task.h"
#include "timers.h"
#include "queue.h"
#include "semphr.h"
#include "event_groups.h"

#include <stdarg.h>
#include <string.h>

/* Variables, Handlers and Buffers */
uint8_t xIsTransmitting = pdFALSE;
uint8_t rxBuffer[SERIALCOMM_RXBUFFERSIZE];

SemaphoreHandle_t srlMutex;
SemaphoreHandle_t rxSemaphore;

static UART_HandleTypeDef UartHandle;

/* Function prototypes */
void prvSerialCommInit(void);
void prvSerialCommDeInit(void);

void vSerialCommTask(void * pvParameters)
{
	uint8_t buff[SERIALCOMM_RXBUFFERSIZE];
	uint32_t length;

	prvSerialCommInit();
	srlMutex = xSemaphoreCreateMutex();
	rxSemaphore = xSemaphoreCreateBinary();

	while(1)
	{
		if (rxSemaphore != NULL)
		{
			if(xSemaphoreTake(rxSemaphore, portMAX_DELAY) == pdTRUE)
			{
				length = SERIALCOMM_TXBUFFERSIZE - SERIALCOMM_USARTx_RX_DMA_INSTANCE->NDTR;
				memcpy(buff, &rxBuffer, length);
				serialIn(buff, length);
				HAL_UART_Receive_DMA(&UartHandle, (uint8_t *)&rxBuffer, SERIALCOMM_RXBUFFERSIZE);
			}
		}
	}

	vTaskDelete(NULL);
}

void printOut (const char *format, ...)
{
	va_list arg;
	char buffer[SERIALCOMM_TXBUFFERSIZE];

	va_start (arg, format);
	vsprintf (buffer, format, arg);
	va_end (arg);

	while(xIsTransmitting == pdTRUE)
	{
		vTaskDelay(1);
	}

	if(__get_IPSR() != 0)
	{
		BaseType_t xHigherPriorityTaskWoken;
		xHigherPriorityTaskWoken = pdFALSE;

		if(xSemaphoreTakeFromISR( srlMutex, &xHigherPriorityTaskWoken ) == pdTRUE)
		{
			HAL_UART_Transmit_DMA( &UartHandle, ( uint8_t * )buffer, strlen( buffer ) );
			xIsTransmitting = pdTRUE;
			xSemaphoreGiveFromISR(srlMutex, &xHigherPriorityTaskWoken);
		}
		portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
	}
	else
	{
		if(xSemaphoreTake( srlMutex, portMAX_DELAY ) == pdTRUE)
		{
			HAL_UART_Transmit_DMA( &UartHandle, ( uint8_t * )buffer, strlen( buffer ) );
			xIsTransmitting = pdTRUE;
			xSemaphoreGive(srlMutex);
		}
	}
}

void serialOut(uint8_t * buffer, uint32_t length)
{
	while(xIsTransmitting == pdTRUE)
	{
		vTaskDelay(1);
	}

	if(__get_IPSR() != 0)
	{
		BaseType_t xHigherPriorityTaskWoken;
		xHigherPriorityTaskWoken = pdFALSE;

		if(xSemaphoreTakeFromISR( srlMutex, &xHigherPriorityTaskWoken ) == pdTRUE)
		{
			HAL_UART_Transmit_DMA( &UartHandle, buffer, length );
			xIsTransmitting = pdTRUE;
			xSemaphoreGiveFromISR(srlMutex, &xHigherPriorityTaskWoken);
		}
		portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
	}
	else
	{
		if(xSemaphoreTake( srlMutex, portMAX_DELAY ) == pdTRUE)
		{
			HAL_UART_Transmit_DMA( &UartHandle, buffer, length );
			xIsTransmitting = pdTRUE;
			xSemaphoreGive(srlMutex);
		}
	}
}

/* By default, implement a simple echo */
__weak void serialIn(uint8_t *buffer, uint32_t length)
{
	buffer[length] = '\0';
	printOut("ECHO: %s\r\n", buffer);
}

void prvSerialCommInit(void)
{
	UartHandle.Instance          = SERIALCOMM_USARTx;
	UartHandle.Init.BaudRate     = SERIALCOMM_USARTx_BAUDRATE;
	UartHandle.Init.WordLength   = UART_WORDLENGTH_8B;
	UartHandle.Init.StopBits     = UART_STOPBITS_1;
	UartHandle.Init.Parity       = UART_PARITY_NONE;
	UartHandle.Init.HwFlowCtl    = UART_HWCONTROL_NONE;
	UartHandle.Init.Mode         = UART_MODE_TX_RX;
	UartHandle.Init.OverSampling = UART_OVERSAMPLING_16;

	if(HAL_UART_Init(&UartHandle) != HAL_OK)
	{
		Error_Handler();
	}

	SET_BIT(SERIALCOMM_USARTx->CR1, USART_CR1_IDLEIE);
	HAL_UART_Receive_DMA(&UartHandle, (uint8_t *)&rxBuffer, SERIALCOMM_RXBUFFERSIZE);
}

void prvSerialCommDeInit(void)
{
	if (HAL_UART_DeInit(&UartHandle) != HAL_OK)
	{
		/* Initialization Error */
		Error_Handler();
	}
}

void HAL_UART_MspInit(UART_HandleTypeDef *huart)
{
	static DMA_HandleTypeDef hdma_tx;
	static DMA_HandleTypeDef hdma_rx;

	GPIO_InitTypeDef  GPIO_InitStruct;

	/*##-1- Enable peripherals and GPIO Clocks #################################*/
	/* Enable GPIO TX/RX clock */
	SERIALCOMM_USARTx_TX_GPIO_CLK_ENABLE();
	SERIALCOMM_USARTx_RX_GPIO_CLK_ENABLE();
	/* Enable USART clock */
	SERIALCOMM_USARTx_CLK_ENABLE();
	/* Enable DMA clock */
	SERIALCOMM_DMAx_CLK_ENABLE();

	/*##-2- Configure peripheral GPIO ##########################################*/
	/* UART TX GPIO pin configuration  */
	GPIO_InitStruct.Pin       = SERIALCOMM_USARTx_TX_PIN;
	GPIO_InitStruct.Mode      = GPIO_MODE_AF_PP;
	GPIO_InitStruct.Pull      = GPIO_NOPULL;
	GPIO_InitStruct.Speed     = GPIO_SPEED_FAST;
	GPIO_InitStruct.Alternate = SERIALCOMM_USARTx_TX_AF;

	HAL_GPIO_Init(SERIALCOMM_USARTx_TX_GPIO_PORT, &GPIO_InitStruct);

	/* UART RX GPIO pin configuration  */
	GPIO_InitStruct.Pin = SERIALCOMM_USARTx_RX_PIN;
	GPIO_InitStruct.Alternate = SERIALCOMM_USARTx_RX_AF;

	HAL_GPIO_Init(SERIALCOMM_USARTx_RX_GPIO_PORT, &GPIO_InitStruct);

	/*##-3- Configure the DMA streams ##########################################*/
	/* Configure the DMA handler for Transmission process */
	hdma_tx.Instance                 = SERIALCOMM_USARTx_TX_DMA_INSTANCE;
#ifdef SERIALCOMM_USARTx_TX_DMA_CHANNEL
	hdma_tx.Init.Channel             = SERIALCOMM_USARTx_TX_DMA_CHANNEL;
#endif
#ifdef SERIALCOMM_USARTx_TX_DMA_REQUEST
	hdma_tx.Init.Request             = SERIALCOMM_USARTx_TX_DMA_REQUEST;
#endif
	hdma_tx.Init.Direction           = DMA_MEMORY_TO_PERIPH;
	hdma_tx.Init.PeriphInc           = DMA_PINC_DISABLE;
	hdma_tx.Init.MemInc              = DMA_MINC_ENABLE;
	hdma_tx.Init.PeriphDataAlignment = DMA_PDATAALIGN_BYTE;
	hdma_tx.Init.MemDataAlignment    = DMA_MDATAALIGN_BYTE;
	hdma_tx.Init.Mode                = DMA_NORMAL;
	hdma_tx.Init.Priority            = DMA_PRIORITY_LOW;

	HAL_DMA_Init(&hdma_tx);

	/* Associate the initialized DMA handle to the the UART handle */
	__HAL_LINKDMA(huart, hdmatx, hdma_tx);

	/* Configure the DMA handler for Transmission process */
	hdma_rx.Instance                 = SERIALCOMM_USARTx_RX_DMA_INSTANCE;
#ifdef SERIALCOMM_USARTx_RX_DMA_CHANNEL
	hdma_rx.Init.Channel             = SERIALCOMM_USARTx_RX_DMA_CHANNEL;
#endif
#ifdef SERIALCOMM_USARTx_RX_DMA_REQUEST
	hdma_rx.Init.Request             = SERIALCOMM_USARTx_RX_DMA_REQUEST;
#endif
	hdma_rx.Init.Direction           = DMA_PERIPH_TO_MEMORY;
	hdma_rx.Init.PeriphInc           = DMA_PINC_DISABLE;
	hdma_rx.Init.MemInc              = DMA_MINC_ENABLE;
	hdma_rx.Init.PeriphDataAlignment = DMA_PDATAALIGN_BYTE;
	hdma_rx.Init.MemDataAlignment    = DMA_MDATAALIGN_BYTE;
	hdma_rx.Init.Mode                = DMA_NORMAL;
	hdma_rx.Init.Priority            = DMA_PRIORITY_HIGH;

	HAL_DMA_Init(&hdma_rx);

	/* Associate the initialized DMA handle to the the UART handle */
	__HAL_LINKDMA(huart, hdmarx, hdma_rx);

	/*##-4- Configure the NVIC for DMA #########################################*/
	/* NVIC configuration for DMA transfer complete interrupt (USARTx_TX) */
	HAL_NVIC_SetPriority(SERIALCOMM_USARTx_DMA_TX_IRQn, configLIBRARY_LOWEST_INTERRUPT_PRIORITY, 1);
	HAL_NVIC_EnableIRQ(SERIALCOMM_USARTx_DMA_TX_IRQn);

	/* NVIC configuration for DMA transfer complete interrupt (USARTx_RX) */
	HAL_NVIC_SetPriority(SERIALCOMM_USARTx_DMA_RX_IRQn, configLIBRARY_LOWEST_INTERRUPT_PRIORITY, 0);
	HAL_NVIC_EnableIRQ(SERIALCOMM_USARTx_DMA_RX_IRQn);

	/* NVIC configuration for USART interrupt */
	HAL_NVIC_SetPriority(SERIALCOMM_USARTx_IRQn, configLIBRARY_LOWEST_INTERRUPT_PRIORITY, 0);
	HAL_NVIC_EnableIRQ(SERIALCOMM_USARTx_IRQn);
}

/**
 * @brief UART MSP De-Initialization
 *        This function frees the hardware resources used in this example:
 *          - Disable the Peripheral's clock
 *          - Revert GPIO, DMA and NVIC configuration to their default state
 * @param huart: UART handle pointer
 * @retval None
 */
void HAL_UART_MspDeInit(UART_HandleTypeDef *huart)
{
	static DMA_HandleTypeDef hdma_tx;
	static DMA_HandleTypeDef hdma_rx;

	/*##-1- Reset peripherals ##################################################*/
	SERIALCOMM_USARTx_FORCE_RESET();
	SERIALCOMM_USARTx_RELEASE_RESET();

	/*##-2- Disable peripherals and GPIO Clocks ################################*/
	/* Configure UART Tx as alternate function  */
	HAL_GPIO_DeInit(SERIALCOMM_USARTx_TX_GPIO_PORT, SERIALCOMM_USARTx_TX_PIN);
	/* Configure UART Rx as alternate function  */
	HAL_GPIO_DeInit(SERIALCOMM_USARTx_RX_GPIO_PORT, SERIALCOMM_USARTx_RX_PIN);

	/*##-3- Disable the DMA Streams ############################################*/
	/* De-Initialize the DMA Stream associate to transmission process */
	HAL_DMA_DeInit(&hdma_tx);
	/* De-Initialize the DMA Stream associate to reception process */
	HAL_DMA_DeInit(&hdma_rx);

	/*##-4- Disable the NVIC for DMA ###########################################*/
	HAL_NVIC_DisableIRQ(SERIALCOMM_USARTx_DMA_TX_IRQn);
	HAL_NVIC_DisableIRQ(SERIALCOMM_USARTx_DMA_RX_IRQn);
}

/**
 * @brief  This function handles DMA RX interrupt request.
 * @param  None
 * @retval None
 */
void SERIALCOMM_USARTx_DMA_RX_IRQHandler(void)
{
	HAL_DMA_IRQHandler(UartHandle.hdmarx);
}

/**
 * @brief  This function handles DMA TX interrupt request.
 * @param  None
 * @retval None
 */
void SERIALCOMM_USARTx_DMA_TX_IRQHandler(void)
{
	HAL_DMA_IRQHandler(UartHandle.hdmatx);
}

/**
 * @brief  This function handles SERIALCOMM_USARTx interrupt request.
 * @param  None
 * @retval None
 */
void SERIALCOMM_USARTx_IRQHandler(void)
{
	/* Check for IDLE flag */
	if (SERIALCOMM_USARTx->SR & USART_SR_IDLE) {         /* We want IDLE flag only */
		/* This part is important */
		/* Clear IDLE flag by reading status register first */
		/* And follow by reading data register */
		volatile uint32_t tmp;                  /* Must be volatile to prevent optimizations */
		tmp = SERIALCOMM_USARTx->SR;                       /* Read status register */
		tmp = SERIALCOMM_USARTx->DR;                       /* Read data register */
		(void)tmp;                              /* Prevent compiler warnings */
		SERIALCOMM_USARTx_RX_DMA_INSTANCE->CR &= ~DMA_SxCR_EN;       /* Disabling DMA will force transfer complete interrupt if enabled */
	}

	HAL_UART_IRQHandler(&UartHandle);
}

/**
 * @brief  Tx Transfer completed callback
 * @param  huart: UART handle.
 * @note   This example shows a simple way to report end of DMA Tx transfer, and
 *         you can add your own implementation.
 * @retval None
 */
void HAL_UART_TxCpltCallback(UART_HandleTypeDef *huart)
{
	xIsTransmitting = pdFALSE;
}

/**
 * @brief  Rx Transfer completed callback
 * @param  huart: UART handle
 * @note   This example shows a simple way to report end of DMA Rx transfer, and
 *         you can add your own implementation.
 * @retval None
 */
void HAL_UART_RxCpltCallback(UART_HandleTypeDef *huart)
{
	BaseType_t xHigherPriorityTaskWoken;
	xHigherPriorityTaskWoken = pdFALSE;
	xSemaphoreGiveFromISR(rxSemaphore, &xHigherPriorityTaskWoken);
	portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
}

/**
 * @brief  UART error callbacks
 * @param  huart: UART handle
 * @note   This example shows a simple way to report transfer error, and you can
 *         add your own implementation.
 * @retval None
 */
void HAL_UART_ErrorCallback(UART_HandleTypeDef *huart)
{
	/* Transfer error in reception/transmission process */
}
