#include "main.h"
#include "stm32f4xx_hal.h"
#include "stm32f4xx_nucleo.h"

#include "FreeRTOS.h"
#include "task.h"
#include "timers.h"
#include "queue.h"
#include "semphr.h"
#include "event_groups.h"

void SystemClock_Config(void);
void USARTConfig(void);

/* Tasks -----------------------------------------------------------*/
void LedBlinkTask(void const * argument);
void SerialPrintTask(void const * argument);
void ButtonHandler(void const * argument);

volatile uint8_t UART_RxBuffer[UART_BufferSize];
volatile uint8_t UART_TxBuffer[UART_BufferSize];
static char dataOut[MAX_BUF_SIZE];
const uint32_t Usart_BaudRate = 115200;
UART_HandleTypeDef UartHandle;

volatile uint32_t counter = 0;
volatile uint32_t toggle = 0;

SemaphoreHandle_t btnSemaphore;

int main(void)
{
  HAL_Init();
  SystemClock_Config();

  BSP_LED_Init(LED2);
  BSP_PB_Init(BUTTON_USER, BUTTON_MODE_EXTI);
  USARTConfig();

  btnSemaphore = xSemaphoreCreateBinary();

  xTaskCreate((void *)LedBlinkTask, "BLINK", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY, NULL);
  xTaskCreate((void *)SerialPrintTask, "SERIAL", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY, NULL);
  xTaskCreate((void *)ButtonHandler, "BUTTON", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY, NULL);

  vTaskStartScheduler();

  while (1)
  {
	  BSP_LED_Toggle(LED2);
	  HAL_Delay(100);
  }

}

void LedBlinkTask(void const * argument)
{
  while(1)
  {
	  BSP_LED_Toggle(LED2);
	  vTaskDelay(1000);
  }
}

void SerialPrintTask(void const * argument)
{
	while(1)
	{
	  counter++;
	  snprintf( dataOut, MAX_BUF_SIZE, "This is a counter: %d\r\n", counter );
	  HAL_UART_Transmit( &UartHandle, ( uint8_t * )dataOut, strlen( dataOut ), 5000 );
	  vTaskDelay(1000);
	}
}

void ButtonHandler(void const * argument)
{
	while(1)
	{
		if (btnSemaphore != NULL)
		{
			if (xSemaphoreTake(btnSemaphore, 1000) == pdTRUE)
		    {
				snprintf( dataOut, MAX_BUF_SIZE, "Button Pressed!\r\n");
				HAL_UART_Transmit( &UartHandle, ( uint8_t * )dataOut, strlen( dataOut ), 5000 );
		    }
		}
	}
}

void HAL_GPIO_EXTI_Callback(uint16_t GPIO_Pin)
{
	BaseType_t xHigherPriorityTaskWoken;
	xHigherPriorityTaskWoken = pdFALSE;

	if(GPIO_Pin == USER_BUTTON_PIN)
	{
		xSemaphoreGiveFromISR(btnSemaphore, &xHigherPriorityTaskWoken);
		portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
	}
}

void SystemClock_Config(void)
{

  RCC_OscInitTypeDef RCC_OscInitStruct;
  RCC_ClkInitTypeDef RCC_ClkInitStruct;

  __HAL_RCC_PWR_CLK_ENABLE();

  __HAL_PWR_VOLTAGESCALING_CONFIG(PWR_REGULATOR_VOLTAGE_SCALE2);

  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSI;
  RCC_OscInitStruct.HSIState = RCC_HSI_ON;
  RCC_OscInitStruct.HSICalibrationValue = 16;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
  RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_HSI;
  RCC_OscInitStruct.PLL.PLLM = 16;
  RCC_OscInitStruct.PLL.PLLN = 336;
  RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV4;
  RCC_OscInitStruct.PLL.PLLQ = 7;
  if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
  {
    _Error_Handler(__FILE__, __LINE__);
  }

  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK|RCC_CLOCKTYPE_SYSCLK
                              |RCC_CLOCKTYPE_PCLK1|RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV2;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;

  if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_2) != HAL_OK)
  {
    _Error_Handler(__FILE__, __LINE__);
  }

  HAL_SYSTICK_Config(HAL_RCC_GetHCLKFreq()/1000);
  HAL_SYSTICK_CLKSourceConfig(SYSTICK_CLKSOURCE_HCLK);
  HAL_NVIC_SetPriority(SysTick_IRQn, 15, 0);
}

void USARTConfig( void )
{

  GPIO_InitTypeDef GPIO_InitStruct;

  USARTx_TX_GPIO_CLK_ENABLE();
  USARTx_RX_GPIO_CLK_ENABLE();
  USARTx_CLK_ENABLE();

  GPIO_InitStruct.Pin       = USARTx_TX_PIN;
  GPIO_InitStruct.Mode      = GPIO_MODE_AF_PP;
  GPIO_InitStruct.Pull      = GPIO_NOPULL;
  GPIO_InitStruct.Speed     = GPIO_SPEED_FREQ_HIGH;
  GPIO_InitStruct.Alternate = USARTx_TX_AF;

  HAL_GPIO_Init(USARTx_TX_GPIO_PORT, &GPIO_InitStruct);

  GPIO_InitStruct.Pin       = USARTx_RX_PIN;
  GPIO_InitStruct.Alternate = USARTx_RX_AF;

  HAL_GPIO_Init( USARTx_RX_GPIO_PORT, &GPIO_InitStruct );

  UartHandle.Instance        = USARTx;
  UartHandle.Init.BaudRate   = Usart_BaudRate;
  UartHandle.Init.WordLength = UART_WORDLENGTH_8B;
  UartHandle.Init.StopBits   = UART_STOPBITS_1;
  UartHandle.Init.Parity     = UART_PARITY_NONE;
  UartHandle.Init.HwFlowCtl  = UART_HWCONTROL_NONE;
  UartHandle.Init.Mode       = UART_MODE_TX_RX;

  if ( HAL_UART_Init( &UartHandle ) != HAL_OK )
  {
    while( 1 );
  }

  UartHandle.pRxBuffPtr = ( uint8_t * )UART_RxBuffer;
  UartHandle.RxXferSize = UART_BufferSize;
  UartHandle.ErrorCode  = HAL_UART_ERROR_NONE;
}

void _Error_Handler(char *file, int line)
{
  while(1)
  {
  }
}

#ifdef  USE_FULL_ASSERT
void assert_failed(uint8_t* file, uint32_t line)
{ 
}
#endif
